# frozen_string_literal: true

module Hermes
  # Il représente une pièce à fournir et contient
  # un statut (pièce manquante, contrôle en cours, pièce rejetée, pièce valide),
  # une liste des codes erreurs remontés,
  # et le lien vers le ou les fichiers.
  #
  # il est lié à un ou plusieurs applicant, ainsi,
  # une pièce peut être retrouvée et resservir pour plusieurs housingRequestFolder
  class RequiredDocument
    include Hermes::Concerns::CommonFieldsConcern
    include Hermes::Concerns::SopraXmlSender
    include Hermes::Concerns::Status
    
    field :nature, type: String # nature de piece à fournir
    field :type, type: String # type de piece à fournir
    field :label, type: String # 
    field :category, type: String
    field :validity_limit, type: Date # Duréle limite de validité , nil= infini
    #field :status, type: String # status de la piece (validée, traitement en cours, erreur)
    field :validation_errors, type: Array # liste de erreurs renvoyées par le KYC
    field :tagset_id, type: String # id de l'espace de travail correspondant
    field :tagsetLocation, type: String # url du serveur sur lequel est stokée le document, nil si c'est en interne
    field :doc_id, type: String
    field :sne_code, type: String
    field :sne_category, type: String
    field :sne_piece_label, type: String
    field :sne_piece_id, type: String
    field :kyc_needed, type: Boolean
    field :folder_where_is_required_for_nud, type: Array
    field :folder_where_is_required_for_consultation, type: Array
    field :folder_where_is_required_for_application, type: Array
    field :last_upload_to_sne_date, type: Date
    field :last_download_from_sne_date, type: Date
    field :last_user_send_file_date, type: DateTime
    field :exists_in_sne, type: Boolean
    field :ged_doc_id, type: String
    field :file_url, type: String
    field :preview_url, type: String
    field :ged_doc_data, type: Hash
    field :metas, type: Hash, default: {}
    field :type_ihm, type: String
    field :ihm_order, type: String
    field :kyc_process, type: String
    field :or_required_document_types, type: Array, default: []
    field :declared, type: Boolean, default: nil
    
    
    belongs_to :applicant, class_name: 'Hermes::Applicant', index: true
    
    writer_auth_roles "collaborateur offre logement", 'superadministrateur'
    
    writer_roles 'Superadmin', 'Administrator'
    reader_roles 'Manager'
    owner_field :user_id
    
    before_create :before_create_action
    before_save :before_save_action
    after_create :after_create_actions
    after_save :after_save_actions
    after_destroy :after_destroy_actions
    
    RequiredDocument.add_status(:single)
    
    common_indexes *%i[
      nature
      type
      category
      validity_limit
      tagset_id
      doc_id
      sne_code
      sne_category
      sne_piece_id
      kyc_needed
    ]
    
    def get_ged_doc_data
      if Rails.env.test?
        data = {
          'data' => {
            'id' => '5e5cd6b677d81d6356ee7edb',
            'type' => 'docs',
            'attributes' => {
              'created_at' => '2020-03-02T10:49:42.778+01:00',
              'updated_at' => '2020-03-02T14:06:04.143+01:00',
              'title' => 'listing type piece justif SNE (1).png',
              'read_level' => 'write',
              'content_type' => 'image/png',
              'tags' => [],
              'percent' => 0,
              'type' => 'file',
              'file_url' => '/send/cf19acd7380b71fd3891b3f5cae8bb43/listing type piece justif SNE 1.png',
              'pages_count' => 1,
              'pages_urls' => [
                '/sassets/408d84b99db067be5232d840dc815c7d/5e5d04bc77d81d5749b24581-5e5d04bc77d81d5749b24582_asset-1583154363-5q1uho76.PNG'
              ],
              'thumbs_urls' => [
                '/sassets/f2ddfe1391e2082d1633740243c31cd4/5e5d04bb77d81d5749b2457d-5e5d04bb77d81d5749b2457e_asset-1583154363-23pojoaj.PNG'
              ],
              'thumb_url' => '/sassets/c8e9ff4c97d12a04b2e258314f6ded5c/5e5d04bb77d81d5749b24575-5e5d04bb77d81d5749b24576_asset-1583154363-gq9dkpft.PNG',
              'preview_url' => '/sassets/2555d9b89590972dd203111da8044517/5e5d04bb77d81d5749b24579-5e5d04bb77d81d5749b2457a_asset-1583154363-jbyqua7v.PNG',
              'responsible' => '5e46c85777d81d3c61114ed1',
              'writers' => [],
              'cycle_id' => '',
              'family_id' => '',
              'file_size' => 374_782,
              'readers' => [],
              'tagset_ids' => [],
              'space_ids' => [],
              'superfields' => {},
              'version_of' => nil,
              'blocs' => [],
              'doc_hidden_part_ids' => []
            },
            'relationships' => {
              'doc_hidden_parts' => {
                'data' => []
              },
              'channel' => {
                'data' => nil
              },
              'instant_messages' => {
                'data' => []
              },
              'spaces' => {
                'data' => [],
                'links' => {
                  'related' => '/api/d2/spaces?doc_id=5e5cd6b677d81d6356ee7edb'
                }
              }
            }
          }
        }
      else
        if ged_doc_data && ged_doc_data['data']['attributes']['pages_urls'] && ged_doc_data['data']['attributes']['pages_urls'].length > 0
          ged_doc_data
        elsif ged_doc_id
          data = get_data_from_ged(ged_doc_id)
          return nil unless data
          
          set(ged_doc_data: data, preview_url: data['data']["attributes"]["preview_url"], updated_at: DateTime.current)
          data
        end
      end
    end
    
    def smartify_urls ged_doc_data
      return {} unless ged_doc_data && ged_doc_data['data'] && ged_doc_data['data']['attributes']
      ged_doc_data['data']['attributes']['pages_urls'].each_with_index do |url, i|
        ged_doc_data['data']['attributes']['pages_urls'][i] = hermes_ged_link(url)
      end
      ged_doc_data['data']['attributes']['thumbs_urls'].each_with_index do |url, i|
        ged_doc_data['data']['attributes']['thumbs_urls'][i] = hermes_ged_link(url)
      end
      ged_doc_data['data']['attributes']['file_url'] = hermes_ged_link(ged_doc_data['data']['attributes']['file_url'])
      ged_doc_data['data']['attributes']['thumb_url'] = hermes_ged_link(ged_doc_data['data']['attributes']['thumb_url'])
      ged_doc_data['data']['attributes']['preview_url'] = hermes_ged_link(ged_doc_data['data']['attributes']['preview_url'])
      ged_doc_data
    end
    
    def force_update_data
      rds = Hermes::RequiredDocument.all
      rds.each do |rd|
        if (rd.ged_doc_id)
          data = rd.get_data_from_ged(rd.ged_doc_id)
          rd.ged_doc_data = data
          rd.file_url = data['data']['attributes']['file_url']
          rd.preview_url = data['data']["attributes"]["preview_url"]
          rd.save
        end
      end
    end
    
    def create_tagset
      Tagset.create(is_space: true, title: applicant.full_name + ' -- ' + type)
    end
    
    def hermes_ged_link(url)
      return "" if !url
      
      "/hermes/gedfile" + url
    end
    
    def create_doc(filename, blob)
      doc = Doc.create(type: 'file', title: filename, tagset_ids: [tagset_id], responsible: user_id)
      doc.blob = blob
      doc.save
      doc
    end
    
    def set_blobs(array_of_blobs)
      # reset du status du document
      self.validation_errors = nil
      self.validity_limit = nil
      self.tagset_id = create_tagset.id
      self.last_user_send_file_date = DateTime.current
      final_blob = format_to_single_pdf(array_of_blobs)
      doc = create_doc(id.to_s + '.pdf', final_blob)
      doc.id.to_s
      self.doc_id = doc.id.to_s
      set_viwed_to_false
    end
    
    def set_viwed_to_false
      folder_where_is_required_for_application.each do |folder_id|
        
        Hermes::HousingRequestFolder.find(folder_id)&.housing_request.set(viewed_by_colaborator_offre_logement: false, updated_at: DateTime.current)
      end
    end
    
    def delete_files
      delete_file_in_ged
      delete_files_in_sne if sne_piece_id
      self.validity_limit = nil
      self.status = "Pièce manquante"
      self.validation_errors = nil
      self.tagset_id = nil
      self.tagsetLocation = nil
      self.doc_id = nil
      self.sne_piece_id = nil
      self.last_upload_to_sne_date = nil
      self.last_download_from_sne_date = nil
      self.last_user_send_file_date = nil
      self.exists_in_sne = false
      self.ged_doc_id = nil
      self.file_url = nil
      self.ged_doc_data = nil
    end
    
    def format_to_single_pdf(array_of_blobs)
      array_of_pdf_blobs = []
      
      pdf_processor = File::Processor::Actions::Image::Pdf.new
      
      array_of_blobs.each do |item|
        if item[:filename].split('.')[item[:filename].split('.').length - 1] == 'pdf'
          array_of_pdf_blobs.push(item[:blob])
        else
          array_of_pdf_blobs.push(pdf_processor.process(item[:blob]))
        end
      end
      blob_concat = nil
      if array_of_pdf_blobs.length > 1
        fp = File::Processor::Actions::Pdf::Concatenate.new
        doc = nil
        
        array_of_pdf_blobs.each do |pdf_blob|
          if doc
            blob_concat = fp.process(pdf_blob)
          else
            doc = create_doc('temporary_blob.pdf', pdf_blob)
            fp.doc_id = doc.id
          end
        end
      else
        blob_concat = array_of_pdf_blobs[0]
      end
      
      blob_concat
    end
    
    #     {
    #       "auteurDepot" => "TEST SOPRA",
    #       "categorieCOD" => "SITPRO",
    #       "categorieLIB" => "Situation professionnelle",
    #       "creationAuteur" => "TEST SOPRA",
    #       "creationDAT" => "2020-02-18T19:01:55.000+01:00",
    #       "dateDepot" => "2020-02-18T19:01:55.000+01:00",
    #       "derniereActionAuteur" => "TEST SOPRA",
    #       "derniereActionDAT" => "2020-02-18T19:01:55.000+01:00",
    #       "derniereActionType" => "Création",
    #       "documentInterne" => "false",
    #       "fonctionnelleDAT" => "2034-02-13T00:00:00.000+01:00",
    #       "idFichier" => "16695",
    #       "idPiece" => "16695",
    #       "indicePertinence" => "Non vérifié",
    #       "nomFichier" => "SNE000185966000025.pdf",
    #       "numeroUnique" => "0130719003263G3U05",
    #       "sensibilite" => "false",
    #       "type" => "Contrat de travail"
    #     },
    def request_from_sne(parsed_xml_sne_piece, unique_number)
      so = Hermes::BoClient::DownLoadPJTask.new(
        num_unique: unique_number,
        pieceID: parsed_xml_sne_piece['idPiece']
      )
      
      if Rails.env == 'test'
        path = [Hermes::Engine.root, %w[spec xml_bouchon], '34.xml'].flatten.join('/').to_s
        pn = Pathname.new(path)
        return {err: 'unknown file'} unless pn.exist?
        
        File.read(pn)
      
      else
        ExecuteSopraTaskJob.perform_now(so.endpoint, so.to_xml, "post", "Hermes::RequiredDocument", id.to_s)
      end
    end
    
    def get_file_from_sne(parsed_xml_sne_piece, unique_number)
      sne_response = request_from_sne(parsed_xml_sne_piece, unique_number)
      return unless sne_response
      res = Hash.from_xml(sne_response.gsub("\n", ''))
      return unless res['FichierPJ']
      set_blobs([{
                   filename: res['FichierPJ']['nomFichier'],
                   blob: res['FichierPJ']['fileContent']
                 }])
      self.last_download_from_sne_date = Date.current
      self.exists_in_sne = true
      save
    end
    
    
    def create_file_in_sne(unique_number, housing_request_id)
      return unless self.doc_id
      so = Hermes::BoClient::CreatePJTask.new(
        housing_request_id: housing_request_id.to_s,
        num_unique: unique_number,
        required_document_id: id.to_s
      )
      # puts so.to_xml
      validate_xml(so.to_xml, 'integrationPJ.xsd')
      res_sne = ExecuteSopraTaskJob.perform_now(so.endpoint, so.to_xml, "post", "Hermes::RequiredDocument", id.to_s)
      
      if (res_sne)
        res = Hash.from_xml(res_sne)
        set(sne_piece_id: res['pieceJustificative']["metadonnees"]["idPiece"], exists_in_sne: true, last_upload_to_sne_date: Date.current, updated_at: DateTime.current, last_xml_validation_report: last_xml_validation_report)
      end
      
    end
    
    def update_file_in_sne(unique_number, housing_request_id)
      so = Hermes::BoClient::UpdatePJTask.new(
        housing_request_id: housing_request_id.to_s,
        num_unique: unique_number,
        required_document_id: id.to_s
      )
      # puts so.to_xml
      validate_xml(so.to_xml, 'integrationPJ.xsd')
      set(last_upload_to_sne_date: Date.current, updated_at: DateTime.current, last_xml_validation_report: last_xml_validation_report)
      
      so.execute
      so
    end
    
    def delete_file_in_sne(unique_number, housing_request_id)
      so = Hermes::BoClient::DeletePJTask.new(
        housing_request_id: housing_request_id.to_s,
        num_unique: unique_number,
        required_document_id: id.to_s
      )
      
      validate_xml(so.to_xml, "supprimerPJ.xsd")
      so.execute
      so
    end
    
    def delete_files_in_sne
      concerned_unique_numbers_and_housing_request_id.each do |unique_numbers_and_housing_request_id|
        delete_file_in_sne(unique_numbers_and_housing_request_id[:unique_number], unique_numbers_and_housing_request_id[:housing_request_id])
      end
    end
    
    def concerned_unique_numbers_and_housing_request_id
      hrs = HousingRequest.any_of({main_applicant_id: applicant&.id}, {main_applicant_id: applicant&.id.to_s}, {applicant_ids: applicant&.id}, applicant_ids: applicant&.id.to_s)
      unique_numbers = []
      hrs.each do |concerned_hr|
        next unless concerned_hr.nud_ids || concerned_hr.nud_ids.empty?
        
        concerned_hr.nuds.each do |nud|
          unique_numbers.push(unique_number: nud.unique_number, housing_request_id: concerned_hr.id)
        end
      end
      unique_numbers.uniq
    end
    
    def reprise_piece_sne_script
      pieces = Hermes::RequiredDocument.where(sne_piece_id: nil,
                                              :sne_code.nin => ["", nil],
                                              :doc_id.nin => ["", nil],
                                              :sne_category.nin => ["", nil],
                                              :sne_piece_label.nin => ["", nil],
                                              :applicant_id.nin => ["", nil]
      )
      pieces.length
      pieces.batch_size(5).each_with_index do |piece, i|
        if (piece.doc_id && piece.applicant)
          piece.upload_file_to_sne
          piece.save
        end
      end
    end
    
    def upload_file_to_sne
      if !Rails.env.test?
        UploadDocInSneJob.perform_later(id.to_s)
      end
    end
    
    def upload_file_to_sne_sync
      if !Rails.env.test?
        UploadDocInSneJob.perform_now(id.to_s)
      end
    end
    
    def after_create_actions;
    end
    
    def update_parent_folders_status
      parent_folder_ids = folder_where_is_required_for_nud + folder_where_is_required_for_consultation + folder_where_is_required_for_application
      parent_folder_ids.each do |folder_id|
        parent_folder = Hermes::HousingRequestFolder.find(folder_id)
        parent_folder.compute_statuses
        parent_folder.save
      end
    end
    
    def set_status
      if declared
        self.status = 'Pièce validée'
      elsif declared === false
          self.status = 'Pièce manquante'
      elsif !doc_id
        self.status = 'Pièce manquante'
      else
        self.status = if kyc_needed
          
                        if (validation_errors.is_a? Array)
                          if validation_errors != []
                            'Pièce invalide'
                          else
                            'A contrôler'
                          end
                        else
                          'Prêt pour traitement'
                        end
                      else
                        'A contrôler'
                      end
      end
    end
    
    def invalid_doc?
      ['pièce invalide', 'pièce rejetée'].include?(status&.downcase)
    end
    
    def missing?
      ['pièce manquante'].include?(status&.downcase)
    end
    
    def delete_file_in_ged
      return if Rails.env.test?
      
      #  DeleteDocInGedJob.perform_later(id.to_s, ged_doc_id.to_s)
    end
    
    def delete_to_ged(ged_doc_id)
      request = Curl::Easy.new(HERMES_CONFIG['ged']['url'] + '/api/d2/docs/' + ged_doc_id) do |request|
        request.ssl_verify_peer = false # (Rails.env == 'production')
        request.ssl_verify_host = false
        request.proxy_url = APP_CONFIG['proxy'] || ENV['http_proxy']
        request.headers['X-Auth-User-Token'] = HERMES_CONFIG['ged']['X-Auth-User-Token']
        request.delete = true
      end
      
      request.perform
      request.status
    end
    
    def post_to_ged(document)
      response = nil
      request = Curl::Easy.new(HERMES_CONFIG['ged']['url'] + '/api/d2/docs')
      
      request.ssl_verify_peer = false # (Rails.env == 'production')
      request.ssl_verify_host=false
      
      request.proxy_url = APP_CONFIG['proxy'] || ENV['http_proxy']
      request.headers['X-Auth-User-Token'] = HERMES_CONFIG['ged']['X-Auth-User-Token']
      if document && file = document.file
        request.multipart_form_post = true
        post_field = Curl::PostField.content('file', document.data)
        post_field.remote_file = document.title.cleaned
        post_field.content_type = document.content_type
        extra_args = [
          post_field,
          Curl::PostField.content('data[attributes][title]', document.title.cleaned)
        ]
      
      end
      
      if (Rails.env == 'test')
        VCR.use_cassette("send_file_to_ged_api_test", :match_requests_on => [:method, :uri, :path]) do
          request.http_post(extra_args)
        end
      else
        request.http_post(extra_args)
      end
      if request.status&.to_s&.match?('2..')
        return JSON.parse(request.body_str)
      else
        Log.perform_later(
                   level: 'error', 
                   domain: 'doc',
                   code: "HERMES_GED_ERR",
                   action: "required_document : #{self.id} -- post_to_ged --",
                   data: {
                     url: HERMES_CONFIG['ged']['url'] + '/api/d2/docs',
                     headers: request.headers,
                     response: WsResponse.from_curl(request).raw_body
                   }.as_json.transform_values{|v| v.to_s})
        return nil
      end
    end
    
    def get_data_from_ged(ged_doc_id_externe)
      request = Curl::Easy.new(HERMES_CONFIG['ged']['url'] + '/api/d2/docs/' + ged_doc_id_externe) do |request|
        request.ssl_verify_peer = false # (Rails.env == 'production')
        request.ssl_verify_host = false
        request.proxy_url = APP_CONFIG['proxy'] || ENV['http_proxy']
        request.headers['X-Auth-User-Token'] = HERMES_CONFIG['ged']['X-Auth-User-Token']
      end
      
      request.perform
      return JSON.parse(request.body_str) if request.status&.to_s&.match?('2..')
    end
    
    def save_file_in_ged
      document = Doc.find(doc_id)
      
      if ged_doc_id
        delete_file_in_ged
        self.ged_doc_id = nil
        self.file_url = nil
        self.ged_doc_data = nil
      end
      data = post_to_ged(document)
      if data
        self.ged_doc_id = data['data']['id']
        self.file_url = data['data']['attributes']['file_url']
        self.ged_doc_data = data
      else
        return false
      end
    end
    
    def start_treatment(k7_name = nil)
      
      if (doc_id && kyc_process)
        process = "::Hermes::Kyc::Kyc#{kyc_process}".constantize.new
        process.start_treatment(self, k7_name)
      end
    end
    
    def try_call_netheos(k7_name = nil)
      if (doc_id && kyc_process)
        
        process = "::Hermes::Kyc::Kyc#{kyc_process}".constantize.new
        process.try_call_netheos(self, k7_name)
      end
    end
    
    def before_create_action
      self.writers.push(applicant.user_id)
    end
    
    def before_save_action
      set_status if validation_errors_changed? || doc_id_changed? || kyc_needed_changed? || declared_changed?
      save_file_in_ged if doc_id_changed? && doc_id
    end
    
    def after_save_actions
      if sne_code && sne_category && doc_id_changed?
        upload_file_to_sne
      end

      check_status_changed
      if status_changed?
        update_parent_folders_status
      end
      update_parent_folders_cache_key
    end
    
    def after_destroy_actions
      delete_file_in_ged if ged_doc_id
      delete_files_in_sne if exists_in_sne && sne_piece_id
      update_parent_folders_cache_key
    end

    def update_parent_folders_cache_key
      parent_folder_ids = folder_where_is_required_for_nud + folder_where_is_required_for_consultation + folder_where_is_required_for_application
      Hermes::HousingRequestFolder.where(:id.in => parent_folder_ids).set(updated_at: Time.now)
    end

    def  status_regarding_validity_limit
      if status == "Pièce validée" && kyc_needed && (validity_limit && DateTime.current > validity_limit)
        "Date de validité dépassée"
      else
        status
      end
      end
    def self.generate_pdf_link(file_url, preview_url)
      coder = HTMLEntities.new
      text = "Lien vers le fichier sur l'espace sécurisé AL'IN"
      html = <<HTML
<!DOCTYPE html>
<html>
<head>
  <title>Document AL&apos;IN : titre</title>
  <meta name="viewport" content="width=device-width,initial-scale=1">
  <style>
  body {
    background-color: #3eb7ca;
    color: #2E2F30;
    text-align: center;
    font-family:  proxima-nova,sans-serif;
    margin: 0;
    padding:0.5em 0 0 0;
  }

  div.dialog {
    width: 75%;
    margin: 2em auto 0;
  }

  div.dialog > div {
    border: 1px solid #CCC;
    border-right-color: #999;
    border-left-color: #999;
    border-bottom-color: #BBB;
    border-top: #B00100 solid 2px;
    background-color: white;
    padding: 7px 5% 0;
    box-shadow: 0 3px 8px rgba(50, 50, 50, 0.17);
  }

  h1 {
    font-size: 100%;
    color: #192c7b;
    line-height: 1.5em;
  }

  div.dialog > p {
    margin: 0 0 1em;
    padding: 1em;
    background-color: #F7F7F7;
    border: 1px solid #CCC;
    border-right-color: #999;
    border-left-color: #999;
    border-bottom-color: #999;
    border-top-color: #d7575c;
    color: #666;
    box-shadow: 0 3px 8px rgba(50, 50, 50, 0.17);
    font-size:0.9em;
  }
  div.dialog > a {
    color: #192c7b;
  }
  div.dialog > p > a > img {
    margin: 10px auto;
    -webkit-box-shadow: 10px 10px 22px -7px rgba(0,0,0,0.75);
    -moz-box-shadow: 10px 10px 22px -7px rgba(0,0,0,0.75);
    box-shadow: 10px 10px 22px -7px rgba(0,0,0,0.75);
  }
  div.footer {
    position:absolute;
    bottom:20px;
    width: 75%;
    text-align: right;
    color: #eee;
    font-size:0.8em;
  }
  div.footer > a {
    text-decoration:none;
    color: #eee;
  }
  </style>
</head>

<body>
  <div class="dialog">
    <div>
      <h1><img src="#{Rails.root}/packages/hermes/config/al-logo.png"><br/>AL&apos;IN Faciliter le logement pour favoriser l&apos;emploi</h1>
    </div>
    <p><a href="#{file_url}">#{coder.encode(text, :named)}<br/>
      <img src="#{preview_url}" max-width="100%"></a></p>
  </div>
  <div class="footer"><a href="http://plugandwork.fr">plugandwork</a></div>
</body>
</html>

HTML
      set_margin = '0'
      kit = PDFKit.new(html, :page_size => 'A4', :margin_top => set_margin + 'in', :margin_right => set_margin + 'in', :margin_bottom => set_margin + 'in', :margin_left => set_margin + 'in')
      kit.to_pdf
    end
  end
end
